#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2023. All Rights Reserved.

from resilient_app_config_plugins.plugin_base import PAMPluginInterface

class MyMockPlugin(PAMPluginInterface):
    def __init__(self, *args, **kwargs):
        pass
    def get(self, key, default=None):
        return "MOCK"
    def selftest(self):
        return True, ""

class MyBadMockPlugin(): # does not implement required Interface
    def __init__(self, *args, **kwargs):
        pass
    def get(self, key):
        return "MOCK"
    def selftest(self):
        return True, ""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2023. All Rights Reserved.

import os
import sys

import pytest
from mock import patch
from tests.shared_mock_data.mock_plugins.mock_plugins import (MyBadMockPlugin,
                                                              MyMockPlugin)

from resilient import constants
from resilient.app_config import AppConfigManager, ProtectedSecretsManager

MOCK_API_KEY_VALUE = "JbkOxTInUg1aIRGxXI8zOG1A25opU39lDKP1_0rfeVQ"

@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY3_VERSION, reason="requires python3.6 or higher")
def test_protected_secrets_manager(fx_write_protected_secrets, fx_reset_environmental_variables):
    os.environ[constants.ENV_VAR_APP_HOST_CONTAINER] = "1"
    path_secrets_dir = fx_write_protected_secrets
    path_jwk_file = os.path.join(path_secrets_dir, ".jwk", "key.jwk")
    psm = ProtectedSecretsManager(path_secrets_dir=path_secrets_dir, path_jwk_file=path_jwk_file)

    key = psm.get("$API_KEY")
    nothing = psm.get("$DOESNT_EXIST")

    assert key == MOCK_API_KEY_VALUE
    assert nothing is None

def test_protected_secrets_manager_PY27(fx_reset_environmental_variables):
    os.environ["UNPROTECTED_API_KEY"] = "passw0rd"
    os.environ["UNPROTECTED_SECRET_WITH_BRACKETS"] = "A[xyz)e}K,/MabS1}:NbJ$("
    psm = ProtectedSecretsManager()

    key = psm.get("$UNPROTECTED_API_KEY")
    key2 = psm.get("$UNPROTECTED_SECRET_WITH_BRACKETS")

    assert key == "passw0rd"
    assert key2 == "A[xyz)e}K,/MabS1}:NbJ$("

def test_app_config_manager_no_plugin():
    original_dict = {
        "a": 1,
        "b": 2,
        "c": 3,
        "d": {
            "a": 1,
            "b": 2
        },
        "nested": {
            "nested": {
                "nested": {
                    "password": "^secret"
                }
            }
        }
    }

    acm = AppConfigManager(original_dict)

    assert isinstance(acm, dict)

    # AppConfigManager objects have to support many ways of attribute accessing
    # for backward compatbility with existing app.config usages
    assert acm["a"] == original_dict["a"]
    assert acm["b"] == original_dict["b"]
    assert acm["c"] == original_dict["c"]
    assert acm.get("a") == original_dict.get("a")
    assert acm.get("b") == original_dict.get("b")
    assert acm.get("c") == original_dict.get("c")
    assert acm.a == original_dict.get("a")
    assert acm.b == original_dict.get("b")
    assert acm.c == original_dict.get("c")
    assert isinstance(acm["d"], AppConfigManager)
    assert isinstance(acm["nested"]["nested"]["nested"], AppConfigManager)
    assert isinstance(acm.get("d"), AppConfigManager)
    assert isinstance(acm.get("nested").get("nested").get("nested"), AppConfigManager)

    assert acm.get("nested").get("nested").get("nested").get("password") == "^secret"

    assert acm._asdict() == acm
    assert acm._asdict().get("nested") == acm.nested

def test_app_config_manager_with_plugin():
    original_dict = {
        "a": 1,
        "b": "^top.level/secret/with_special.characters",
        "d": {
            "a": "$IN_ENV",
            "b": "^shhh"
        },
        "nested": {
            "nested": {
                "nested": {
                    "password": "^secret",
                    "password2": "^{secret1} and ^{secret1}"
                }
            }
        }
    }

    acm = AppConfigManager(original_dict, pam_plugin_type=MyMockPlugin)

    assert acm.get("b") == "MOCK"
    assert acm.get("nested").get("nested").get("nested").get("password") == "MOCK"
    assert acm.get("nested").get("nested").get("nested").get("password2") == "MOCK and MOCK"
    assert acm["d"]["a"] == "$IN_ENV"

    os.environ["IN_ENV"] = "FOUND"
    assert acm["d"]["a"] == "FOUND"

    # check that __repr__ is not substituting in the protected values
    assert "^secret" in repr(acm)
    assert "^{secret1} and ^{secret1}" in repr(acm)
    assert "MOCK" not in repr(acm)

@pytest.mark.parametrize("item, prefix, expected", [
    # basic, old usage
    ("^item in keyring", "^", "MOCK"),

    # nothing substituted
    ("nothing", "^", "nothing"),
    ("nothing", "^{", "nothing"),

    # mid-item substituting
    ("^{item}some {text} in $between ^{another item}", "^{", "MOCKsome {text} in $between MOCK"),
    ("text ^{thensomestuff}", "^{", "text MOCK")
])
def test_replace_secret_in_config_pam_plugin(item, prefix, expected):
    manager = MyMockPlugin()
    replaced = AppConfigManager.replace_secret_in_config(item, manager, prefix)

    assert replaced == expected

@pytest.mark.parametrize("item, prefix, secret, expected", [
    ("$SECRET", "$", "A[xyz)e}K,/MabS1}:NbJ$(}$", "A[xyz)e}K,/MabS1}:NbJ$(}$"),
    ("${SECRET}${SECRET_2}", "${", "A[xyz)e}K,/MabS1}:NbJ$(}${", "A[xyz)e}K,/MabS1}:NbJ$(}${2")
])
def test_replace_secret_in_config_protected_secret_env_only(item, prefix, secret, expected, fx_reset_environmental_variables, caplog):
    manager = ProtectedSecretsManager()
    os.environ["SECRET"] = secret
    os.environ["SECRET_2"] = "2"

    replaced = AppConfigManager.replace_secret_in_config(item, manager, prefix)

    assert replaced == expected
    assert "Failed to find" not in caplog.text


@pytest.mark.skipif(sys.version_info < constants.MIN_SUPPORTED_PY3_VERSION, reason="requires python3.6 or higher")
def test_replace_secret_in_config_protected_secret(fx_write_protected_secrets, fx_reset_environmental_variables):
    os.environ[constants.ENV_VAR_APP_HOST_CONTAINER] = "1"
    path_secrets_dir = fx_write_protected_secrets
    path_jwk_file = os.path.join(path_secrets_dir, ".jwk", "key.jwk")
    manager = ProtectedSecretsManager(path_secrets_dir=path_secrets_dir, path_jwk_file=path_jwk_file)

    # NOTE: couldn't parametrize this test because of the protected secrets directory
    # creation not playing well with the parametrized mark
    for item, prefix, expected in [
        ("$protected_secret", "$", "FOUND"),

        # nothing substituted
        ("no protected secret", "$", "no protected secret"),
        ("no protected secret", "${", "no protected secret"),

        # mid-item substitutions
        ("${item} with other stuff here", "${", "FOUND with other stuff here"),
        ("start ${item} end", "${", "start FOUND end"),
        ("start ${item} another ${API_KEY}", "${", "start FOUND another {0}".format(MOCK_API_KEY_VALUE))
    ]:
        with patch("resilient.app_config.helpers.get_config_from_env") as patch_env_get:
            patch_env_get.return_value = "FOUND"
            replaced = AppConfigManager.replace_secret_in_config(item, manager, prefix)

            assert replaced == expected

def test_app_config_manager_bad_plugin():
    with pytest.raises(ValueError) as captured_err:
        AppConfigManager({}, MyBadMockPlugin)

    assert "'pam_plugin_type' must be a subclass of 'PAMPluginInterface'" in str(captured_err)



# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2023. All Rights Reserved.

"""Command-line argument parser for Resilient apps"""

import argparse
import getpass
import logging
import os
import sys

from resilient import constants, helpers
from resilient.app_config import AppConfigManager, ProtectedSecretsManager

if sys.version_info.major == 2:
    from io import open

    from resilient import (ensure_unicode, get_proxy_dict,
                           get_resilient_circuits_version)
else:
    from resilient.co3 import (ensure_unicode, get_proxy_dict,
                               get_resilient_circuits_version)

try:
    # For all python < 3.2
    import backports.configparser as configparser
except ImportError:
    import configparser

logger = logging.getLogger(__name__)


class ArgumentParser(argparse.ArgumentParser):
    """Helper to parse common Resilient command line arguments.

    Optionally, arguments can be specified in a config file, section :samp:`[resilient]`.

    It is expected that a command line utility that needs to work with a Resilient
    server will create its own class that inherits this class.  Its `__init__`
    method should call :samp:`self.add_argument(...)` as necessary.

    See https://docs.python.org/3/library/argparse.html for more details.
    """

    DEFAULT_PORT = 443
    config = None

    def getopt(self, section, opt):
        """Get a single option value.

        :param section: The configuration section.
        :param opt: The configuration option.
        :return: The value, or None if not present.
        """
        if self.config:
            if opt in self.config.options(section):
                return self.config.get(section, opt)
        return None

    def getopts(self, section, opt):
        """Get a comma-separated option value as an array.

        :param section: The configuration section.
        :param opt: The configuration option.
        :return: The value, or [] if not present.
        """
        if self.config:
            if opt in self.config.options(section):
                return self.config.get(section, opt).split(u",")
        return []

    def __init__(self, config_file=None):
        super(ArgumentParser, self).__init__()

        # Read configuration options.
        if config_file:
            config_path = ensure_unicode(config_file)
            config_path = os.path.expanduser(config_path)
            if os.path.exists(config_path):
                try:
                    self.config = configparser.ConfigParser(interpolation=None)
                    with open(config_path, 'r', encoding='utf-8') as f:
                        first_byte = f.read(1)
                        if first_byte != u'\ufeff':
                            # Not a BOM, no need to skip first byte
                            f.seek(0)
                        self.config.read_file(f)
                except Exception as exc:
                    logger.warn(u"Couldn't read config file '%s': %s", config_path, exc)
                    self.config = None
            else:
                logger.warn(u"Couldn't read config file '%s'", config_file)

        default_email = self.getopt("resilient", "email")
        default_password = self.getopt("resilient", "password")
        default_key_id = self.getopt("resilient", "api_key_id")
        default_key_secret = self.getopt("resilient", "api_key_secret")
        default_host = self.getopt("resilient", "host")
        default_port = self.getopt("resilient", "port") or self.DEFAULT_PORT
        default_proxy = self.getopts("resilient", "proxy")
        default_org = self.getopt("resilient", "org")
        default_cafile = self.getopt("resilient", "cafile")
        default_cache_ttl = int(self.getopt("resilient", "cache_ttl") or 0)
        default_proxy_host = self.getopt("resilient", "proxy_host")
        default_proxy_port = self.getopt("resilient", "proxy_port") or 0
        default_proxy_user = self.getopt("resilient", "proxy_user")
        default_proxy_password = self.getopt("resilient", "proxy_password")
        default_stomp_prefetch_limit = int(self.getopt("resilient", "stomp_prefetch_limit") or 20)
        default_resilient_mock = self.getopt("resilient", "resilient_mock")

        default_max_request_retries = self.getopt(constants.PACKAGE_NAME, constants.APP_CONFIG_REQUEST_MAX_RETRIES) or constants.APP_CONFIG_REQUEST_MAX_RETRIES_DEFAULT
        default_request_retry_delay = self.getopt(constants.PACKAGE_NAME, constants.APP_CONFIG_REQUEST_RETRY_DELAY) or constants.APP_CONFIG_REQUEST_RETRY_DELAY_DEFAULT
        default_request_retry_backoff = self.getopt(constants.PACKAGE_NAME, constants.APP_CONFIG_REQUEST_RETRY_BACKOFF) or constants.APP_CONFIG_REQUEST_RETRY_BACKOFF_DEFAULT

        # PAM plugin configurations
        if helpers.is_running_in_app_host(constants.ENV_VAR_APP_HOST_CONTAINER):
            # no default in app host
            default_pam_type = None
        else:
            # integration server defaults to Keyring
            default_pam_type = self.getopt(constants.PACKAGE_NAME, constants.PAM_TYPE_CONFIG) or constants.PAM_DEFAULT_PAM_TYPE

        self.add_argument("--email",
                          default=default_email,
                          help="The email address to use to authenticate to the Resilient server.")

        self.add_argument("--password",
                          default=default_password,
                          help="WARNING:  This is an insecure option since the password "
                               "will be visible to other processes and stored in the "
                               "command history.  The password to use to authenticate "
                               "to the Resilient server.  If omitted, the you will be prompted.")

        self.add_argument("--api_key_id",
                          default=default_key_id,
                          help="The api key id for API key.")

        self.add_argument("--api_key_secret",
                          default=default_key_secret,
                          help="WARNING:  This is an insecure option since the key secret "
                               "will be visible to other processes and stored in the "
                               "command history.  The password to use to authenticate "
                               "to the Resilient server.  If omitted, the you will be prompted.")

        self.add_argument("--host",
                          default=default_host,
                          required=default_host is None,
                          help="Resilient server host name.")

        self.add_argument("--port",
                          type=int,
                          default=default_port,
                          help="Resilient server REST API port number.")

        self.add_argument("--proxy",
                          default=default_proxy,
                          nargs="*",
                          help="An optional HTTP proxy to use when connecting.")

        self.add_argument("--org",
                          default=default_org,
                          help="The name of the organization to use.  If you are a member "
                               "of just one organization, then you can omit this argument.")

        self.add_argument("--cafile",
                          default=default_cafile,
                          help="The name of a file that contains trusted certificates.")

        self.add_argument("--cache-ttl",
                          default=default_cache_ttl or 240,
                          type=int,
                          help="TTL for API responses when using cached_get")

        self.add_argument("--proxy_host",
                          default=default_proxy_host,
                          help="HTTP Proxy host for Resilient Connection.")

        self.add_argument("--proxy_port",
                          type=int,
                          default=default_proxy_port,
                          help="HTTP Proxy port for Resilient Connection.")

        self.add_argument("--proxy_user",
                          default=default_proxy_user,
                          help="HTTP Proxy username for Resilient connection authentication.")

        self.add_argument("--proxy_password",
                          default=default_proxy_password,
                          help="HTTP Proxy password for Resilient connection authentication.")

        self.add_argument("--stomp-prefetch-limit",
                          default=default_stomp_prefetch_limit,
                          type=int,
                          help="MAX number of Action Module messages to send before ACK is required")

        self.add_argument("--{0}".format(constants.APP_CONFIG_REQUEST_MAX_RETRIES),
                          type=int,
                          default=default_max_request_retries,
                          help="Max number of times to retry a request to SOAR before exiting. Defaults to 5")

        self.add_argument("--{0}".format(constants.APP_CONFIG_REQUEST_RETRY_DELAY),
                          type=int,
                          default=default_request_retry_delay,
                          help="Number of seconds to wait between retries. Defaults to 2")

        self.add_argument("--{0}".format(constants.APP_CONFIG_REQUEST_RETRY_BACKOFF),
                          type=int,
                          default=default_request_retry_backoff,
                          help="Multiplier applied to delay between retry attempts. Defaults to 2")

        self.add_argument("--{0}".format(constants.PAM_TYPE_CONFIG),
                          default=default_pam_type,
                          help="PAM plugin type to use for pulling secrets. Defaults to Keyring")

        v_resc = get_resilient_circuits_version()

        # Having --resilient-mock here allows us to run unit tests for resilient-circuits and resilient-sdk
        # This argument will be added if:
        # - resilient-circuits is not installed
        # - resilient-circuits is installed and is > 34
        if not v_resc or (v_resc and v_resc.get("major") > 34):
            self.add_argument("--resilient-mock",
                            default=default_resilient_mock, 
                            help="<path_to_mock_module>.NameOfMockClass")

    def parse_args(self, args=None, namespace=None, ALLOW_UNRECOGNIZED=False):
        """
        Parse the configuration options and command-line arguments.

        :return: Note: the return value is a dict, not a Namespace.
        """

        if ALLOW_UNRECOGNIZED or constants.ALLOW_UNRECOGNIZED:
            return self.parse_known_args(args, namespace)[0]

        # (the implementation calls parse_known_args)
        args = super(ArgumentParser, self).parse_args(args, namespace)
        return args

    def parse_known_args(self, args=None, namespace=None):
        """
        Parse the configuration options and command-line arguments.

        :return: Note: the return value is a dict, not a Namespace.
        """
        # Parse the known arguments
        args, argv = super(ArgumentParser, self).parse_known_args(args, namespace)
        return _post_process_args(args), argv


def _post_process_args(args):
    # Post-process any options that reference keyring or environment variables
    args = AppConfigManager(vars(args))

    # Post-processing for other special options
    password = args.password
    email = args.email

    while (not password and email) and (not args.get("no_prompt_password")):
        password = getpass.getpass()
    args["password"] = ensure_unicode(password)

    if args.get("cafile"):
        args["cafile"] = os.path.expanduser(args.cafile)

    if args.get("stomp_cafile"):
        args["stomp_cafile"] = os.path.expanduser(args.stomp_cafile)

    if args.get("proxy_host"):
        args["proxy"] = get_proxy_dict(args)

    return args


def parse_parameters(options):
    """
    Given a dict that has configuration keys mapped to values,
       - If a value begins with '^', redirect to fetch the value from
         the secret key stored in the plugin provided by pam_type.
       - If a value begins with '$', fetch the value from environment or protected secret
    """

    plugin_type_str = helpers.get_pam_type_name(options, ProtectedSecretsManager())
    plugin_type = helpers.load_pam_plugin(plugin_type_str) if plugin_type_str else None

    # though options is already an AppConfigManager object, the plugin_type was not available
    # when the options object was initially created -- it was created with the default value.
    # we need to recreate it with the new pam plugin
    options = AppConfigManager(options, pam_plugin_type=plugin_type)

    return options



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2023. All Rights Reserved.

import logging

from resilient_app_config_plugins.constants import (
    PAM_SECRET_PREFIX, PAM_SECRET_PREFIX_WITH_BRACKET)
from resilient_app_config_plugins.plugin_base import PAMPluginInterface
from six import string_types

from resilient import constants, helpers

LOG = logging.getLogger(__name__)

class ConfigDict(dict):
    """A dictionary, with property-based accessor

    >>> opts = {"one": 1}
    >>> cd = ConfigDict(opts)
    >>> cd["one"]
    1
    >>> cd.one
    1

    """
    def __getattr__(self, name):
        """Attributes are made accessible as properties"""
        try:
            return self[name]
        except KeyError:
            raise AttributeError()

class ProtectedSecretsManager:
    """
    Class to manage all protected secrets and env variables.
    Works with existing protected secrets logic introduced in v47.0.

    If running in integration server, only allows for use of ENV variables
    """

    def __init__(self, path_secrets_dir=constants.PATH_SECRETS_DIR, path_jwk_file=constants.PATH_JWK_FILE):
        self.data = {}

        # find all protected secrets in the PATH_SECRETS_DIR folder
        # and retain the decryption key and encrypted value in self.data
        if helpers.is_running_in_app_host(constants.ENV_VAR_APP_HOST_CONTAINER):
            for config_key in helpers.get_protected_secrets_keys(path_secrets_dir):
                if helpers.protected_secret_exists(config_key, path_secrets_dir, path_jwk_file):
                    # protected secrets are stored as tuple of token, key
                    self.data[config_key] = helpers.get_protected_secret_token_and_key(config_key, path_secrets_dir, path_jwk_file)

    def get(self, config_key, default=None):
        """
        Read key from data.
        If found in data, decrypt using the stored token, key pair.
        If not found, then should be looked for in ENV

        :param config_key: key to retrieve value of
        :type config_key: str
        :return: secret value of config_key either decrypted from protected secret or from ENV
        :rtype: str
        """
        if not config_key:
            return default

        config_key = config_key.lstrip(constants.PROTECTED_SECRET_PREFIX)

        if config_key in self.data:
            tkn, key = self.data.get(config_key)

            protected_secret = helpers.decrypt_protected_secret(tkn, key, config_key)
            protected_secret = protected_secret if protected_secret else helpers.get_config_from_env(config_key, default)
        else:
            protected_secret = helpers.get_config_from_env(config_key, default)

        return protected_secret

class AppConfigManager(ConfigDict):
    """
    Intermediary to manage plain-text configs, $-prefixed protected secrets/env vars,
    and ^-prefixed PAM secrets.

    The AppConfigManager class manages all of those, with late-binding using the
    __getattr__ method to govern when data is retrieved from sensitive locations.

    A ProtectedSecretManager is required to manage all protected secrets and a
    PAM plugin *should* be provided to manage externally managed credentials.
    If no plugin is provided, plain text values will be returned for ^.

    :param dict: initial dictionary, defaults to None
    :type dict: dict, optional
    :param pam_plugin_type: PAM plugin to manage external credentials, defaults to None
    :type pam_plugin_type: ``resilient_app_config_plugins.plugin_base.PAMPluginInterface``, optional
    """

    # static protected secrets manager variable
    # used for pulling protected secrets here and in
    # any plugins that need access to protected secrets
    protected_secrets_manager = ProtectedSecretsManager()

    def __init__(self, dict={}, pam_plugin_type=None, key="_"):
        super(AppConfigManager, self).__init__(dict)

        if pam_plugin_type:
            if not issubclass(pam_plugin_type, PAMPluginInterface):
                raise ValueError("'pam_plugin_type' must be a subclass of 'PAMPluginInterface'")
            self.pam_plugin = pam_plugin_type(protected_secrets_manager=AppConfigManager.protected_secrets_manager, key=key)
        else:
            self.pam_plugin = None

        self._convert_sub_dicts_to_obj(pam_plugin_type)

    ########################################
    ##### superclass 'dict' overrides ######
    ########################################

    def get(self, key, default=None):
        # must override the default 'get' because we need it to use the
        # custom '__getitem__' which will use the PAM logic when necessary
        try:
            return self[key]
        except KeyError:
            return default
        
    def __repr__(self):
        # overriding the default '__repr__' will allow
        # for hiding the credentials that are meant to be hidden
        data = {}
        for key in self:
            # have to bypass this class' logic which incorporates
            # pam and protected secret substitution
            data[key] = super(AppConfigManager, self).__getitem__(key)
        return str(data)

    def __getitem__(self, key):
        """
        This is the key part -- overwrite the behavior of the dictionary
        for retrieving items. The dictionary's item is retrieved. This item
        may or may not start with a protected prefix. If it does, use either the
        protected secrets manager or the pam secrets manager is to resolve the
        item that is to be returned

        NOTE: either ``get()`` methods called here could throw errors. That is
        by design that we don't trap them as we want those managers themselves to
        be able to communicate to the original caller of this method.

        :param key: key to search for in the underlying dict
        :type key: object
        :return: item from the dictionary, protected secret, or the pam plugin
        """
        item = super(AppConfigManager, self).__getitem__(key)

        # NOTE: the order of these checks is important -- the check for the version
        # with bracket has to be checked first, as otherwise the bracket won't be
        # removed from the item
        # also NOTE: there can be a mix of ^{} and ${} items in the secret or any
        # number of {} secrets in the configs

        # handle protected secrets
        if isinstance(item, string_types) and constants.PROTECTED_SECRET_PREFIX_WITH_BRACKET in item:
            item = self.replace_secret_in_config(item, AppConfigManager.protected_secrets_manager, constants.PROTECTED_SECRET_PREFIX_WITH_BRACKET)
        elif isinstance(item, string_types) and item.startswith(constants.PROTECTED_SECRET_PREFIX):
            item = self.replace_secret_in_config(item, AppConfigManager.protected_secrets_manager, constants.PROTECTED_SECRET_PREFIX)

        # handle pam plugin secrets
        if self.pam_plugin and isinstance(item, string_types) and PAM_SECRET_PREFIX_WITH_BRACKET in item:
            item = self.replace_secret_in_config(item, self.pam_plugin, PAM_SECRET_PREFIX_WITH_BRACKET)
        elif self.pam_plugin and isinstance(item, string_types) and item.startswith(PAM_SECRET_PREFIX):
            item = self.replace_secret_in_config(item, self.pam_plugin, PAM_SECRET_PREFIX)

        return item

    def _convert_sub_dicts_to_obj(self, plugin_type):
        """
        Recursively create AppConfigManagers for any sub-dictionaries of the top-level
        dict. Using the underlying dict (which is iterable via ``self``), we can
        iterate over the data of the underlying dictionary.

        Plugin type is not persisted for the whole AppConfigManager class -- thus it
        must be provided here to propagate to sub-dictionaries.

        :param plugin_type: plugin type for the app config manager
        :type plugin_type: ``resilient_app_config_plugins.plugin_base.PAMPluginInterface``
        """
        for key in self:
            if isinstance(self[key], dict):
                # for deeper levels of dictionaries, pass through their "key"
                # so that future uses might maintain that information
                self[key] = AppConfigManager(dict=self[key], pam_plugin_type=plugin_type, key=key)

    def _asdict(self):
        """
        This method is needed to maintain backwards compatibility with
        the way that AppFunctionComponent.app_configs used to work.

        :return: simply returns itself as AppConfigs behave like namedtuples and dicts
        :rtype: AppConfig
        """
        return self


    @staticmethod
    def replace_secret_in_config(item, secret_manager, secret_prefix):
        """
        Use either protected secrets or app config plugin to substitute
        in values where appropriate for item from app.config

        The method here is mostly required to support having multiple
        secrets in a config. This method deals with multiple substitutions
        and, in conjunction with the logic in __getitem__, will substitute
        all secrets in a config.

        To have multiple secrets in a config, brackets must be used.

        **Example:**

        Assuming that you have protected secrets TABLE_SECRET_NAME
        and VALUE_SECRET_NAME, a sql query could be constructed in
        an app.config like this:

        .. code-block::

            [fn_my_app]
            sql_query=FROM ${TABLE_SECRET_NAME} select ${VALUE_SECRET_NAME};

        :param item: value from app.config
        :type item: str
        :param secret_manager: Protected Secrets Manager or App Config Plugin
        :type secret_manager: ProtectedSecretsManager|PAMPluginInterface
        :param secret_prefix: "$" or "${" or "^" or "^{"
        :type secret_prefix: str
        :return: substituted value of app.config item
        :rtype: str
        """
        original_item = item
        if secret_prefix not in item:
            return item
        start = item.index(secret_prefix)

        # loop through any prefixed values in the item and substitute them
        # the loop here is required for multiple values in a line
        while start < len(item):
            end = item.index("}", start) if "}" in item[start:] else len(item)
            prefixed_item = item[start:end+1]
            unprefixed_item = prefixed_item.replace(secret_prefix, "").replace("}", "")
            secret_value = secret_manager.get(unprefixed_item, prefixed_item)

            # since we set prefixed_item to be the default value that will
            # be returned when no value is found, log a warning telling the
            # user that their value wasn't found
            if secret_value == prefixed_item:
                manager_type_str = "PAM Plugin" if isinstance(secret_manager, PAMPluginInterface) else "Protected Secrets"
                LOG.warning("Failed to find '%s' in %s. To properly deploy your app, please make sure '%s' is correctly configured",
                          prefixed_item, manager_type_str, prefixed_item)
            else:
                LOG.debug("Substituting value for '%s' in %s", prefixed_item, original_item)

            item = u"{0}{1}{2}".format(item[0:start], secret_value, item[end+1:])

            # restart the search from the one beyond where we just subbed in the found value
            length_secret = len(secret_value) if secret_value else 0
            start = item.index(secret_prefix, start+length_secret) if secret_prefix in item[start+length_secret:] else len(item)
        return item


